#include "stdafx.h"
#include <sys/types.h>
#ifndef _WINDOWS
#include <sys/socket.h>
#include <sys/uio.h>
#include <pthread.h>
#include <signal.h>
#else
#include "wincompat.h"
#endif
#if !defined(_MSC_EXTENSIONS) && !defined(__MINGW32_VERSION)
#include <unistd.h>
#endif
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include "pdictclient.h"
#include "utils.h"

#ifndef _WINDOWS
static int should_exit = 0;
static int changes = 0;
pdict_reason_t expected_event_type = PDR_ENTRY_ADDED;

static void
sigalrm(int sig)
{
	should_exit = 1;
}
#endif

static int event_handled = 0;
static pthread_cond_t event_handled_cv;
static pthread_mutex_t event_lock;

static void
event_handler(const char *key, const char *val, pdict_reason_t r, void *v)
{
	if (strcmp(key, "foo") == 0 && strcmp(val, "/a\as\\x99df.0") == 0 && r
	    == expected_event_type) {
		event_handled = 1;
		pthread_cond_signal(&event_handled_cv);
	}
}

void
error(const char *errdesc, void *arg)
{
	printf("\nerror: %s\n", errdesc);
	abort();
}

static int
test_escaping(const unsigned char *src)
{
	extern int escape(const char *src, char **dstp);
	char *escaped, *unescaped;
	int res;

	assert(escape((const char *)src, &escaped));
	assert(unescape(escaped, &unescaped));
	res = strcmp((char *)unescaped, (char *)src);

	if (res != 0) {
		int i;
		fprintf(stderr, "while working with %s -> %s...\n", src,
		    escaped);
		for (i = 0; i < strlen((char *)src); i++)
			if (!unescaped[i])
				break;
			else if (unescaped[i] != src[i])
				fprintf(stderr,
				    "unescaped[%d] differs: 0x%x vs src 0x%x\n",
				    i, unescaped[i], src[i]);
		if (strlen((char *)src) != strlen((char *)unescaped))
			fprintf(stderr, "src len %d vs unescaped len %d\n",
			    (int)strlen((char *)src),
			    (int)strlen((char *)unescaped));
	}
	free((void *)escaped);
	free((void *)unescaped);

	return res == 0;
}

int
main(const int argc, const char *argv[])
{
	const char *host;
	const char *port;
	char errdesc[1024];
	pdc_session_t *pdcs;
#ifndef _WINDOWS
	const int period = 1;
#endif
	int fd;
	pdc_listen_id_t id;

#ifdef _WINDOWS
	{
		WSADATA wsad;
		WORD wsav;

		wsav = MAKEWORD(2, 2);
		WSAStartup(wsav, &wsad);
	}
#endif

	assert(test_escaping((unsigned char *)"asdf"));
	assert(test_escaping((unsigned char *)"/asdf"));
	assert(test_escaping((unsigned char *)"/asdf.0"));
	assert(test_escaping((unsigned char *)"/a\asdf.0"));
	assert(test_escaping((unsigned char *)"/a\as\\df.0"));
	assert(test_escaping((unsigned char *)"/a\as\\x99df.0"));
	{
		unsigned char *foo = malloc(20);
		assert(foo);
		strcpy((char *)foo, "/a\as-qdf.0");
		foo[4] = 0x99;
		assert(test_escaping(foo));
		free(foo);
	}
	pthread_cond_init(&event_handled_cv, NULL);
	pthread_mutex_init(&event_lock, NULL);
	if (!pdc_init()) {
		perror("pdc_init");
		return 1;
	}

#if defined(USE_ARGS)
	if (argc != 3) {
		printf("pdictclientexample <dst-hostname/ip> <dst-port>\n");
		return 1;
	}
	host = argv[1];
	port = argv[2];
#else
	host = "127.0.0.1";
	port = "5001";
#endif
	if (!stream_server_connect(host, port, &fd, errdesc,
	    sizeof (errdesc))) {
		fprintf(stderr, "connect(%s:%s): %s\n", host, port,
		    errdesc);
		return 1;
	}
	if (!(pdcs = pdc_session_alloc(fd, pu_read, fd, pu_write, pu_close, 0,
	    0, 0))) {
		perror("pdc_session_alloc");
		fflush(stderr);
		return 1;
	}

	if (!(id = pdc_listen(pdcs, ".*", event_handler, NULL, errdesc,
	    sizeof (errdesc))) || (!pdc_enable_periodic_reports(pdcs, 1,
	    errdesc, sizeof (errdesc))) || (pdc_remove(pdcs, "foo", errdesc,
	    sizeof (errdesc)), !pdc_set(pdcs, "foo", "/a\as\\x99df.0", 1,
	    errdesc, sizeof (errdesc)))) {
		fprintf(stderr, "pdc_listen: %s\n", errdesc);
		fflush(stderr);
		return 1;
	}
	pdc_flush(pdcs, errdesc, sizeof (errdesc));
	assert(event_handled);
	event_handled = 0;
	expected_event_type = PDR_VALUE_CHANGED;
	assert(pdc_set(pdcs, "foo", "/a\as\\x99df.0", 0, errdesc,
	    sizeof (errdesc)));
	printf("waiting for event to be handled (testing threading)...\n");
	while (!event_handled) {
		assert(!pthread_mutex_lock(&event_lock));
		assert(!pthread_cond_wait(&event_handled_cv, &event_lock));
		assert(!pthread_mutex_unlock(&event_lock));
	}
	event_handled = 0;
	pdc_async_set(pdcs, "foo", "/a\as\\x99df.0", 0, error, NULL);
	printf("waiting for event to be handled (testing async set)...\n");
	while (!event_handled) {
		pthread_mutex_lock(&event_lock);
		pthread_cond_wait(&event_handled_cv, &event_lock);
		pthread_mutex_unlock(&event_lock);
	}
#ifndef _WINDOWS
	signal(SIGALRM, sigalrm);
	alarm(period);
	while (!should_exit) {
		pdc_async_set(pdcs, "foo", "bar", 0, error, NULL);
		changes++;
	}
	printf("%d changes in %ds\n", changes, period);
#else
	printf("ok!\n");
#endif
	assert(pdc_ignore(pdcs, id, errdesc, sizeof (errdesc)));
	assert(pdc_quit(pdcs, errdesc, sizeof (errdesc)));

	exit(0);
}
